;*******************************************************************************
;
;                               COS-CALC.INC
;
;       141 Byte (Real mode) cosine table calculation for 386+ CPU's.
;
;                         Coding by Olivier Reubens
;                  (a.k.a. Tasmaniac / (CiD / ôypernov)
;
;              Released to the public domain on August 5, 1994
;
;
;    Use as you like.  But at least send me some greets if you use it :)
;
;*******************************************************************************
;
; Following macro will calculate a cosine table of [0, /2] (90).
; Calculation is performed in fixed point (24 bit precision) integer math.
;
; The resulting values are NOT perfect.  The 24 bit fixed point result will
; be off by a maximum of 3 (1.78814E-07 or 0.000000178814).  If you
; absolutely need PERFECT 24 bit fixed point values or require more bits of
; precision (not very likely) you will have to precalculate or have an
; extended table calculation routine.
;
; Calculation of the cosine is based on the formula by McLaurin.
;
;                     2    4    6                  2*n
;                    x    x    x              n   x
; cos (x)      = 1 -  +  -  + ... + (-1) * 
;                    2!   4!   6!                (2*n)!
; x in radials
;
;
; The divisor ((2*n)!) will become big very fast.  After only 6 generations
; the result will become so high that another generation would overflow a
; 32bit integer value.  This is fortunate since it makes the generation much
; easier, AND results in very fast, accurate results.
; As a drawback, this is also unfortunate.  To correctly calculate a cosine
; for a complete 0 to 2 (360) you would about 13 or 14 generations.  This
; may seem simple, but the truth is different.  For 13 generations, the
; divisor will be 26! (4.032914611E+26).  You would need to store temporary
; results with 90bits (or more) of precision.  Clearly an almost impossible
; task.
; There are two alternatives though. The first consists of using a quarter
; table, and on each sine/cosine required, adjust accordingly (code for this
; is also provided.
; The other alternative would be to calculate a quarter table and complete
; the rest by copying appropriate values to the rest of the table.  You
; could do this easily with only 30 to 40 bytes extra code.
; Choice one offers less memory use at a SLIGHT penalty in speed.  Choice
; two offers speediest sines/cosines at the cost of memory.
;
;*******************************************************************************


;-------------------------------------------------------------------------------
; Calc_CosTab  Degrees
;
; Input :
;  ES:DI (or ES:EDI in a 32bit segment) pointing to the start of the cosine
;        table.
;
;  Degree : Any expression that evaluates to a non-zero integral number.
;           <Degrees> may have any non-zero value fitting in a 16 bit register.
;           <Degrees> indicates the number of 'steps' in /2 (90).
;           Please note, the macro calculates the table from element 0 up
;           to AND including the degree specified.  As a result, if you
;           wanted a cosine table in 1 degree steps you would specify 90.
;           and have a table of 91 (!!) elements.
;
;-------------------------------------------------------------------------------
;
; BFP means Bits fixed point.  24BFP thus means that the value is a fixed
; point integer value with 24 bits of decimal precision.
;
;-------------------------------------------------------------------------------
MACRO         Calc_CosTab  Degrees
              CLD                      ; Increment on String instructions.

              MOV    EBP,0             ; Degree Loop counter.
         ; Calculate Cosine for 'BP' Degrees.
Calc_Cos:     PUSH   EBP               ; Save EBP
              PUSH   EDI               ; Save EDI
         ; Convert 'Degree' in BP to radial value.
              MOV    EAX,EBP           ; Degree in EAX for Mul.
              MOV    EDX,421657428     ; PI/2 in 28BFP
              MUL    EDX               ; EBP*(Pi/2) (28BFP)
              MOV    ECX,Degrees       ; Number of steps in Pi/2 (90)
              DIV    ECX               ; EBP*(Pi/2)/Degrees (28BFP) or radial
                                       ; value of the EBP degree (=angle)
         ; Since the multiplication value for the generation is allways
         ; an even power of the angle value.  We'll precalculate that angle
         ; value once, and reuse it in every generation.
              MUL    EAX               ; EDX:EAX = angle in 56BFP or
                                       ; EDX = angle in 24BFP
              MOV    EDI,EDX           ; EDI = angle in 24BFP

              MOV    EBP,1000000h      ; 1.0 in 24BFP,  EBP is the accumulated
                                       ; value through each generation of
                                       ; the formula.
              MOV    ECX,EBP           ; 1.0 in 24BFP, ECX is the
                                       ; multiplication value for each
                                       ; generation of the formula (= x)
              MOV    ESI,1             ; Fractional part for each generation
                                       ; of the formula. (= 2n!)
              MOV    BL,1              ; Value for fractional multiplicator
              MOV    BH,6              ; Number of generations
Generation:
         ; Calculate Fractional part.
              MOV    EAX,ESI           ; EAX = current fractional value
              MOVZX  EDX,BL            ; EDX = n
              MUL    EDX               ; EAX * n
              INC    BL                ; n+1
              MOV    DL,BL             ; MOVZX EDX,BL, but EDX is zero. this
                                       ; is 2 bytes smaller
              MUL    EDX               ; EAX * n * (n+1)
              INC    BL                ; n+2 (ready for next generation)
              MOV    ESI,EAX           ; ESI = new fractional value
         ; Calculate Multiplier
              MOV    EAX,ECX           ; Multiplier in 24BFP
              MUL    EDI               ; * angle (24BFP) resulting in a
                                       ; 48BFP value in EDX:EAX
              SHRD   EAX,EDX,24        ; SHR  EDX:EAX,24  !! EDX unchanged !!
              MOV    ECX,EAX           ; ECX = new multiplier value.
         ; Multiplier/Fractional
              XOR    EDX,EDX           ; EDX = 0 for DIV
              DIV    ESI               ; Divide by 2n!
              SUB    EBP,EAX           ; Add to accumulated value
              NEG    EBP               ; 'fake' the (-1) value.  This will
                                       ; work for an EVEN number of generations
                                       ; otherwise an aditional NEG EBP is
                                       ; needed at the end.

              DEC    BH                ; One generation done.
              JNZ    Generation        ; More ?
              MOV    EAX,EBP
              POP    EDI
              POP    EBP
;*******************************************************************************
;  Store value in table.  modify as needed.
;*******************************************************************************
              SHR    EAX,9
              CMP    AX,08000h
              JNE    NoAdjust
              MOV    AX,07FFFh
NoAdjust:     STOSW
;*******************************************************************************
              INC    EBP
              CMP    BP,Degrees
              JBE    Calc_Cos

ENDM          Calc_CosTab

